/*
 * element.cpp: definition of Element class
 */

#include <math.h>
#include <windows.h>
#include <GL/gl.h>

#include "point.h"
#include "element.h"

#define BOUNDARY_VALUE -0.5

/*
 * construction
 */
Element::Element()
: u(0.0), v(0.0),
  leftElemPtr(0), rightElemPtr(0), aboveElemPtr(0), belowElemPtr(0),
  result(0.0)
{
  // the array p will be created with all points (0.0, 0.0) by default - see 'point.h'
}

Element::Element(Point centre)
: u(0.0), v(0.0),
  leftElemPtr(0), rightElemPtr(0), aboveElemPtr(0), belowElemPtr(0),
  result(0.0)
{
  // create vertices in anti-clockwise order (square has side len = 1.0)
  p[0] = Point(centre.x - 0.5, centre.y - 0.5);
  p[1] = Point(centre.x + 0.5, centre.y - 0.5);
  p[2] = Point(centre.x + 0.5, centre.y + 0.5);
  p[3] = Point(centre.x - 0.5, centre.y + 0.5);
}

/*
 * add pointers to neighbours (or 0 if no neighbour)
 */
void Element::addNeighbours(Element* _leftElemPtr, Element* _rightElemPtr, Element* _aboveElemPtr, Element* _belowElemPtr)
{
  leftElemPtr  = _leftElemPtr;
  rightElemPtr = _rightElemPtr;
  aboveElemPtr = _aboveElemPtr;
  belowElemPtr = _belowElemPtr;
}

/*
 * set u and v values
 */
void Element::init(double _u, double _v)
{
  u = _u;
  v = _v;

  result = u + v;
}

/*
 * do some sort of calculation involving the neightbours
 */
void Element::update()
{
  double ul, ur, va, vb;

  ul = ur = va = vb = 0.0;

  // because this is a method of the element class, and we are dealing
  // with pointers to Element objects, we can access their private data
  ul = leftElemPtr  ? leftElemPtr->u  : BOUNDARY_VALUE;
  ur = rightElemPtr ? rightElemPtr->u : BOUNDARY_VALUE;
  va = aboveElemPtr ? aboveElemPtr->v : BOUNDARY_VALUE;
  vb = belowElemPtr ? belowElemPtr->v : BOUNDARY_VALUE;

  // 'u' and 'v belong to 'this' instance of the element class
  u = u + pow(2.0, sin(ul*ul - va*va));
  v = v + cos(ur*ur + vb*vb);
}

/*
 * normalise u and v against the length of the vector (u,v)
 */
void Element::normalise()
{
  double m = length();

  // note: no check for divide by zero here
  u /= m;
  v /= m;
}

/*
 * return the length of the vector represented by (u,v)
 */
double Element::length()
{
  return sqrt(u*u + v*v);
}

/*
 * return the sum of u and v
 */
double Element::sumOfParts()
{
  result = u + v;

  return result;
}

/*
 * draw this element
 */
void Element::draw(GLubyte* colour)
{
  glColor4ubv(colour);

  glBegin(GL_POLYGON);
  for (int i=0; i<4; i++) {
    glVertex2d(p[i].x, p[i].y);
  }
  glEnd();
}

/*
 * E N D  O F  F I L E
 */
