/*
 * grid.cpp: definition of the Grid class
 */

#include <stdlib.h>
#include <iostream.h>
#include <strstrea.h>
#include <fstream.h>
#include <windows.h>
#include <GL/gl.h>

#include "point.h"
#include "element.h"
#include "index.h"
#include "randnum.h"
#include "grid.h"

#define COLOUR_TABLE_SIZE 320
#include "defaultMap.h"

/*
 * generate a 2D array of elements
 */
bool Grid::generateArray()
{
  // grab memory
  elem = new Element[idx.size()];
  if (!elem) {
    return false;
  }

  // overwrite empty elements with those containing corner points
  for (int i=0; i<ncols; i++) {
    for (int j=0; j<nrows; j++) {
      elem[idx(i, j)] = Element(Point(double(i) + 0.5, double(j) + 0.5));
    }
  }

  return true;
}

/*
 * add the addresses of neighbouring elements to each element
 * in the array - use 0 for address if at boundary
 */
void Grid::addNeighbours()
{
  // the idx object is of class Index and checks if we are 
  // asking for a position within the grid boundaries
  Index idx(ncols, nrows);

  Element *leftPtr, *rightPtr, *abovePtr, *belowPtr;

  for (int i=0; i<ncols; i++) {
    for (int j=0; j<nrows; j++) {
      leftPtr =  idx.isOK(i-1, j) ? &elem[idx(i-1, j)] : 0;
      rightPtr = idx.isOK(i+1, j) ? &elem[idx(i+1, j)] : 0;
      abovePtr = idx.isOK(i, j-1) ? &elem[idx(i, j-1)] : 0;
      belowPtr = idx.isOK(i, j+1) ? &elem[idx(i, j+1)] : 0;
      
      elem[idx(i, j)].addNeighbours(leftPtr, rightPtr, abovePtr, belowPtr);
    }
  }
}

/*
 * Construction & Destruction
 */
Grid::Grid()
: ncols(-1), nrows(-1), ntimeSteps(-1), currentStep(0), elem(0), isDrawing(false)
{
}

Grid::Grid(int _ncols, int _nrows, int _ntimeSteps)
: ncols(_ncols), nrows(_nrows), ntimeSteps(_ntimeSteps), currentStep(0),
  idx(_ncols, _nrows), elem(0), isDrawing(false)
{
  // generate the element array
  if (!generateArray()) {
    cerr << "Out of memory!\n";
    exit(-2);
  }

  // add the addresses of each element's neighbours
  addNeighbours();
}

/*
 * initialise the (u,v) values in each element
 * do it randomly here, but could be read in from file
 */
void Grid::init()
{
  // initialise the min/max result values
  minResult = 1.0e20;
  maxResult = -1.0e20;

  // set number of timesteps done back to zero
  currentStep = 0;

  // r is an instance of my class RandNum that seeds the random
  // number generator with the current time and then returns
  // a series of (pseudo) random numbers. Number return is between 0 and 1
  RandNum r;

  for (int i=0; i<ncols; i++) {
    for (int j=0; j<nrows; j++) {
      elem[idx(i, j)].init(r.next(), r.next());
      // compute the result and test against the min/max values of results
      double test = elem[idx(i, j)].result;
      if (test < minResult) {
        minResult = test;
      }
      else if (test > maxResult) {
        maxResult = test;
      }
    }
  }
}

/*
 * delete the memory used by grid - don't call any methods of Grid after calling this method!
 */
void Grid::destroy()
{
  if (elem) {
    delete[] elem;
  }

  // set elem to zero so if we call a Grid method we get a clean crash
  elem = 0;
}

/*
 * do a timestep by calling the 'update' method on each element
 * and them normalising the vector represented by (u,v)
 */
void Grid::step()
{
  // initialise the min/max result values
  minResult = 1.0e20;
  maxResult = -1.0e20;

  int i, j;

  if (currentStep < ntimeSteps) {
    // do the calculation
    for (i=0; i<ncols; i++) {
      for (j=0; j<nrows; j++) {
        elem[idx(i, j)].update();
      }
    }
    // normalise vector (u,v)
    for (i=0; i<ncols; i++) {
      for (j=0; j<nrows; j++) {
        Element* e = &elem[idx(i, j)];
        e->normalise();
        // compute the result and test against the min/max values of results
        double test = e->sumOfParts();
        if (test < minResult) {
          minResult = test;
        }
        else if (test > maxResult) {
          maxResult = test;
        }
      }
    }
    currentStep++;
  }
}

/*
 * have we done all the time steps?
 */
bool Grid::nextStep()
{
  return currentStep < ntimeSteps;
}

/*
 * output the magnitudes
 */
void Grid::output()
{
  char outName[128];
  strstream str(outName, 128, ios::out);

  str << "grid" << currentStep << ".dat" << '\0';

  ofstream out(outName);

  for (int i=0; i<ncols; i++) {
    for (int j=0; j<nrows; j++) {
      out << elem[idx(i, j)].result << " ";
    }
    out << "\n";
  }
  
  out.close();
}

/*
 * draw the grid using OpenGL
 */
void Grid::draw()
{
  isDrawing = true;

  // generate colours
  double diffResult = maxResult - minResult;

  for (int i=0; i<ncols; i++) {
    for (int j=0; j<nrows; j++) {
      Element* e = &elem[idx(i, j)];
      double ratio = (e->result - minResult) / diffResult;
      int colourTableIndex = int(ratio * (COLOUR_TABLE_SIZE-1));
      e->draw(defaultMap[colourTableIndex]);
    }
  }

  isDrawing = false;
}

/*
 * E N D  O F  F I L E
 */
